#!/usr/bin/perl
#
# callmanager sales stats
#
# $Id: soldstat.cgi,v 1.11 2000/10/24 22:24:14 tlinden Exp tlinden $
#
# render statistik image based on database data and user input for style!
#
# Copyright (c) 2000 ConSol* GmbH
# Author: Thomas Linden <tom@consol.de>
#
# the CGI takes the following parameters:
#
# state: render     - actually render the image
#        showedit   - display db edit screen, all data on one page!
#        showeditsingle - display edit screen for ona one item.(requires idx)
#        edit       - modify data
#        new        - create new db item
#        del        - delete an db item
# ==> edit and new requires the following fields set:
# idx, month, day, year, time, salesman, customer, product, volume, note
# for the new method all time values are not required and will be set to the current time.
# the del method requires just the idx of the db item.
#
# type:  month      - image of current(or specified) month.
#        quarter    - image of current(or specified) quarter
#        year       - image of the current(or specified) year
#
# Output manipulation:
#
# month:      1 - 12
# year:       \d\d\d\d (i.e. 2000)
# quarter:    1,2,3,4
# width:      \d+ (pixel)
# height:     \d+ (pixel)
# ticks:      \d+ (number of lines in X-Axis)
# values:     0 | 1 (0: show values, 1: do not show values)
# factor:     \d+ (divide values by this number, i.e. / 1000, the X-Label will show it too)
# volspace:   \d+ (will be added to the highest value for getting space at the upper side)
# version:    0 (if set, the name and the actually version will be added to the title, 
#                0 turns it off, the default is 1!)
# return:     image | html (specifies, what soldstat will return to the client, default: image)
#
# debug:      1 (if set to 1, then various data will be dumed out instead of an image)
# dumpspot:   1 (if set, the GD::Graph::get_hotspot return value will be printed out)
#
# the script requires a file called rgb.txt, which holds color definitions needed by GD!

use CGI;
use CGI::Carp qw(fatalsToBrowser);
use Fcntl ':flock';
use DBI;
use strict;
use Data::Dumper;
use GD;
use GD::Graph::lines;
use GD::Graph::linespoints;
use GD::Graph::mixed;
use GD::Graph::colour;
use GD::Graph::Data;


sub getdate;
sub show_enterscreen;
sub start_html;
sub end_html;
sub render;
sub show_error;
sub show_editscreen;
sub del;


my(
   $dbdriver, $dbname, $dbhost, $dbuser, $dbpasswd, $database, $DB, $lock, $tp_count, $highest,
   $baseurl, $basepath, $imgpath, $me, $vol_space,
   $RGB, @size_month, @size_year, @size_quarter, %MonthName,
   $query, %input, $maintitle, $table, $colwidth, %HEX_COLORS, $GdColor,
   $LOGO, $POSITION, $y_max_value, $imghttp,
   $Xwidth, $Ywidth,
   $config,
  );


#
# get the config
#
$config = "/opt/httpd/cgi-bin/soldstat.rc";

if (-e $config) {
  eval `cat $config`;
}
else {
  die "Could not read $config!\n";
}



#
# calculate version number
my ($Version, $Revision, $NULL);
$Revision = "";
$Version = "$Revision: 1.11 $NULL";
$Version =~ s/: //;
$Version =~ s/\s*$//;


# establish database connection
&sql_connect;


# get CGI input
$query = new CGI;

# store CGI input
foreach ($query->param) {
  $input{$_} = $query->param($_);
}

#
# main decision making loop
#
if ($input{state} eq "") {
  #
  # starting, just display the data enter screen and exit
  #
  &show_enterscreen;
}
elsif ($input{state} eq "render") {
  #
  # render the image and display it
  #
  &render;
}
elsif ($input{state} eq "showedit") {
  #
  # display the edit screen for data editing
  #
  &show_editscreen( $input{order} );
}
elsif ($input{state} eq "showeditsingle") {
  #
  # display the edit screen for data editing
  #
  &show_editsingle( $input{idx} );
}
elsif ($input{state} eq "edit") {
  #
  # save edited data
  #
  &save("edit");
}
elsif ($input{state} eq "new") {
  #
  # save a new data entry
  #
  &save("new");
}
elsif ($input{state} eq "del") {
  #
  # delete an entry
  #
  &del;
}
else {
  #
  # display a default error screen and exit
  #
  &show_error("unknown option");
}

#
# always exit successful
#
exit(0);

END {
  #
  # always do this after logout
  #
  if ($DB) {
    my $unlock = $DB->prepare("UNLOCK TABLES") || die $DB->errstr;
    $unlock->execute() || die $DB->errstr();
    $DB->disconnect || die $DB->errstr;
  }
}


sub show_enterscreen {
  #
  # starting, just display the data enter screen and exit
  #
  &start_html($maintitle);
  print qq(
<div align=center>
<form name="$me" method="POST">
<input type=hidden name=state value="new">
<table bgcolor="#ccccff" border=0>
<tr>
<td colspan=2>
);
print &menu;
print qq(
</td></tr>
<tr>
  <td colspan=2 align=center><h3>Please enter the CallManager Sales volume</h3></td>
</tr>
<tr>
  <td align=right>Your Name</td><td><input type=text name=salesman size=$colwidth></td>
</tr>
<tr>
  <td align=right>Customer</td><td><input type=text name=customer size=$colwidth></td>
</tr>
<tr>
  <td align=right>Product</td><td><input type=text name=product size=$colwidth></td>
</tr>
<tr>
  <td align=right>Sales volume (DM)</td><td><input type=text name=volume size=$colwidth></td>
</tr>
<tr>
  <td align=right>Optional note</td><td><input type=text name=note size=$colwidth></td>
</tr>
<tr>
  <td align=center><input type=submit name=submit value="save"></td>
  <td align=center><input type=reset name=reset value="reset"></td>
</tr>
<tr>
<td colspan=2>
);
print &menu;
print qq(
</td></tr>
</table>
</div>
  );
  &end_html;
}


sub show_editsingle {
  #
  # starting, just display the data enter screen and exit
  #
  my($idx) = @_;

  my %data = &sql_all_raw($idx);

  &start_html($maintitle);
  print qq(
<div align=center>
<form name="$me" method="POST">
<input type=hidden name=state value="edit">
<input type=hidden name=idx value="$idx">
<table bgcolor="#ccccff" border=0>
<tr>
<td colspan=2>
);
print &menu;
print qq(
</td></tr>
<tr>
  <td colspan=2 align=center><h3>edit sales database item</h3></td>
</tr>
<tr>
<td align=right>Time</td><td>
<input size=2 type=text name="day" value="$data{day}">.
<input size=2 type=text name="month" value="$data{month}">.
<input size=4 type=text name="year" value="$data{year}">
<input size=10 type=text name="time" value="$data{time}">
</td>
</tr>
<tr>
  <td align=right>Salie</td><td><input type=text name=salesman size=$colwidth value="$data{salesman}"></td>
</tr>
<tr>
  <td align=right>Customer</td><td><input type=text name=customer size=$colwidth value="$data{customer}"></td>
</tr>
<tr>
  <td align=right>Product</td><td><input type=text name=product size=$colwidth value="$data{product}"></td>
</tr>
<tr>
  <td align=right>Sales volume (DM)</td><td><input type=text name=volume size=$colwidth value="$data{volume}"></td>
</tr>
<tr>
  <td align=right>Optional note</td><td><input type=text name=note size=$colwidth value="$data{note}"></td>
</tr>
<tr>
  <td align=center><input type=submit name=submit value="save"></td>
  <td align=center><input type=reset name=reset value="reset"></td>
</tr>
<tr>
<td colspan=2>
);
print &menu;
print qq(
</td></tr>
</table>
</div>
  );
  &end_html;
}



sub start_html {
  #
  # will be used by every html printing sub
  #
  my($pagetitle) = @_;
  print "content-type: text/html\n\n\n";
  print qq(
         <!doctype html public "-//W3C//DTD HTML 4.0 Transitional//EN">
	 <html>
	 <head>
	    <meta http-equiv="Content-Type" content="text/html; charset=iso-8859-1">
            <meta name="Author" content="Thomas Linden, soldstat.cgi">
            <title>$pagetitle</title>
         </head>
	 <body bgcolor="#FFFFFF" link="#ffffff" vlink="#ffffff">
  );
}


sub end_html {
  #
  # will be used by every html printing sub
  #
  print qq(
</body>
</html>
  );
}


sub menu {
  #
  # navigation "bar" on top and bottom of edit(*) screen(s)
  #
  my $menu = '
<table width="100%" bgcolor="#c4c4c4" border="0" cellspacing="0" cellpadding="2">
<tr>
<td>
 <table width="100%" bgcolor="#f4f4f4" border="0" cellspacing="0" cellpadding="1">
   <tr>
   <td bgcolor="#c4c4c4" align="left">';
   $menu .= "<b><a href=\"$baseurl/$me\">new entry</a> | "
         .  "<a href=\"$baseurl/$me?state=showedit\">edit database</a> | "
         .  "<a href=\"$baseurl/$me?state=render&type=month\">monthly statistic</a> | "
         .  "<a href=\"$baseurl/$me?state=render&type=quarter\">quarterly statistic</a> | "
         .  "<a href=\"$baseurl/$me?state=render&type=year\">yearly statistic</a></b>";
  $menu .= '
   </td>
   </tr>
 </table>
</td>
</tr>
</table>
';
  return $menu;
}


sub show_editscreen {
  #
  # get data and pass it through a form to the user
  #
  my($order) = @_;
  $order = "idx" if(!$order);

  my(%data, $idx, $month, $day, $year, $time, $salesman, $customer, $product, $volume, $note);

  my $baselink = "$baseurl/$me?state=showedit&order";

  &start_html($maintitle);

  print qq(
<div align=center>
<table bgcolor="#ccccff" border=0>
<tr>
<td colspan=12>);
print &menu;
print qq(
</td>
</tr>
<tr>
  <td colspan=12 align=center><h3>edit sales database</h3></td>
</tr>
<tr bgcolor="#c4c4c4">
<td><b><a href="$baselink=idx">idx</a></b></td>
<td><b><a href="$baselink=month">month</a></b></td>
<td><b><a href="$baselink=day">day</a></b></td>
<td><b><a href="$baselink=year">year</a></b></td>
<td><b><a href="$baselink=time">time</a></b></td>
<td><b><a href="$baselink=salesman">salesman</a></b></td>
<td><b><a href="$baselink=customer">customer</a></b></td>
<td><b><a href="$baselink=product">product</a></b></td>
<td><b><a href="$baselink=volume">volume</a></b></td>
<td><b><a href="$baselink=note">note</a></b></td>
<td>&nbsp</td><td>&nbsp</td>
</tr>
  );


  %data = &sql_all_raw;

  foreach my $idx (sort { $data{$a}->{$order} <=> $data{$b}->{$order} } keys %data) {
    print qq(
       <tr>
       <form name="$me.$idx" method="POST">
       <input type=hidden name=state value="edit">

          <td width="0*"><input size=4 type=text name="idx" value="$data{$idx}->{idx}"></td>
          <td width="0*"><input size=2 type=text name="month" value="$data{$idx}->{month}"></td>
          <td width="0*"><input size=2 type=text name="day" value="$data{$idx}->{day}"></td>
          <td width="0*"><input size=4 type=text name="year" value="$data{$idx}->{year}"></td>
          <td width="0*"><input size=10 type=text name="time" value="$data{$idx}->{time}"></td>
          <td width="0*"><input size=10 type=text name="salesman" value="$data{$idx}->{salesman}"></td>
          <td width="0*"><input size=20 type=text name="customer" value="$data{$idx}->{customer}"></td>
          <td width="0*"><input size=5 type=text name="product" value="$data{$idx}->{product}"></td>
          <td width="0*"><input size=10 type=text name="volume" value="$data{$idx}->{volume}"></td>
          <td width="0*"><input size=20 type=text name="note" value="$data{$idx}->{note}"></td>
          <td width="0*"><input type=submit name=submit value="change"></td>
          <td width="0*" bgcolor="#c4c4c4"><a href="$baseurl/$me?state=del&idx=$idx"><b>del</b></a></td>
       </form>
       </tr>
    );
  }
print qq(
<tr>
<td colspan=12>
);

print &menu;

print qq(
</td>
</tr>
</table>
  );
  &end_html;
}



sub save {
  #
  # save entry(s)
  #
  my($mode) = @_;

  my $insert = "INSERT INTO $table SET "
        . "idx=0, "
	. "month=?, "
	. "day=?, "
	. "year=?, "
	. "time=?, "
	. "salesman=?, "
	. "customer=?, "
	. "product=?, "
	. "volume=?, "
	. "note=?";

  my $update = "UPDATE $table SET "
	. "month=?, "
	. "day=?, "
	. "year=?, "
	. "time=?, "
	. "salesman=?, "
	. "customer=?, "
	. "product=?, "
	. "volume=?, "
	. "note=? "
	. "WHERE idx = ?";


  if ($input{salesman} eq "" || $input{customer} eq "" || $input{product} eq "" || $input{volume} eq "") {
    &show_error("All fields must be filled out!");
    exit(0);
  }


  if ($mode eq "new") {
    my($day, $month, $year, $time) = &getdate;
    my $statement = $DB->prepare($insert) || die $DB->errstr();
    $statement->execute(
			$month, $day, $year, $time,
			$input{salesman}, $input{customer},
			$input{product}, $input{volume},
			$input{note}
		       ) || die $DB->errstr();
    &show_enterscreen;
  }
  elsif ($mode eq "edit") {
    my $statement = $DB->prepare($update) || die $DB->errstr();
    $statement->execute(
			$input{month}, $input{day}, $input{year}, $input{time},
			$input{salesman}, $input{customer},
			$input{product}, $input{volume},
			$input{note}, $input{idx}
		       ) || die $DB->errstr();
    &show_editscreen("");
  }
  # everything ok until here.
}


sub del {
  #
  # delete an entry
  #
  my $sql = "DELETE FROM $table WHERE idx = ?";
  my $statement = $DB->prepare($sql) || die $DB->errstr();
  $statement->execute($input{idx}) || die $DB->errstr();
  &show_editscreen;
}


sub getdate
{
  #
  # return well formatted date+time
  #
  my($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime(time);
  $year += 1900;
  $mon += 1;
  $mon =~ s/^(\d)$/0$1/;
  $hour =~ s/^(\d)$/0$1/;
  $min =~ s/^(\d)$/0$1/;
  $sec =~ s/^(\d)$/0$1/;
  $mday =~ s/^(\d)$/0$1/;
  return $mday, $mon, $year, "$hour:$min:$sec";
}



sub render {
  #
  # render the graphic and print it out
  #
  my(%data);
  my(
     $quarter, $type, $min_month, $min_year,
     $max_volume, $width, $height, $x_tick_offset, $x_label_skip, $x_label,
     $title, $show_values, %volume, @days, @y_data, @values, $next_first_volume,
     $y_tick_number, $y_label_skip, $maxvolume, $TATES, $x_tick_number, $x_long_ticks, $x_tick_length,
     $graph, $x_min_value, $x_max_value, $y_min_value, $x_labels_vertical, @links, $img_name
    );
  if($input{debug}) {
    &start_html;
    print "<pre>";
  }

  %data = &sql_all;

  $vol_space    = ($input{volspace}) ? $input{volspace} : $vol_space;

  #print Dumper(\%data) if($input{debug});

  my($day, $month, $year, $time) = &getdate;
  $month = ($input{month}) ? $input{month} : $month;
  $year  = ($input{year})  ? $input{year}  : $year;

  $type = $input{type};

  if ($input{quarter}) {
    $quarter = $input{quarter};
  }
  else {
    if ($month <= 12 ) {
      $quarter = 4;
    }
    if ($month <= 9) {
      $quarter = 3;
    }
    if ($month <= 6) {
      $quarter = 2
    }
    else {
      $quarter = 1;
    }
  }

  #
  # type specific decision
  # valid types are: month, quarter, year
  #
  if ($type eq "month") {
    ######################### MONTH #########################
    my ($last_char, $new_int, $i, $y_max_value2, $new1, $new2,$y_tick_number);
    if ($month == 1) {
      $min_month = 12;
      $min_year = $year - 1;
    }
    else {
      $min_month = $month - 1;
      $min_year = $year;
    }

    if ($input{nosum}) {
      $max_volume = &get_last_volume($month, $year);
      $y_min_value = 0;
    }
    else {
      $max_volume = &get_last_max_volume($min_month, $min_year);
      #$y_min_value = ($max_volume) ? &rounddown($max_volume) : 0;
      $y_min_value = $max_volume;
    }


    $next_first_volume = &get_next_first_volume($month + 1, $year);

    ($width, $height) = @size_month;
    if($input{width} && $input{height}) {
        $width = $input{width};
        $height = $input{height};
    }

    $graph = new GD::Graph::linespoints($width, $height);
    #$graph = new GD::Graph::mixed($width, $height);

    $x_tick_offset = 1; # 1. tick bei erstem Tag nach der ersten Woche (huh?)
    $x_label_skip  = 1; # alle 7 Tage
    $x_label       = $MonthName{$month} . " $year";
    $x_long_ticks  = 1;
    $x_labels_vertical = 0;
    $title         = $maintitle;
    $show_values   = 1;
    foreach (sort { $data{$a}->{day} <=> $data{$b}->{day} } keys %data) {
      # get sales volumes
      next if($data{$_}->{year} != $year || $data{$_}->{month} != $month);
      $volume{$data{$_}->{day}} = $data{$_}->{volume}; # i.e. %volume = ( 1 => 10000, 4 => 34000 );
      $volume{data}->{$data{$_}->{day}} = $data{$_};
    }
    @days          = &num_days($year, $month);
    $x_min_value = 1;
    $x_max_value = $#days;

    # add volume to each next day and so forth...
    foreach (@days) {
      # iterate over %volume using @days
      if ($volume{$_} == 0) {
	if ($_ == 1 && $max_volume != 0) {
	  $volume{$_} = $max_volume;
	}
	elsif ($_ == 1 && $max_volume == 0) {
	  $volume{$_} = 0;
	}
	elsif ($_ == ($#days + 1) && $next_first_volume > 0) {
	  $volume{$_} = $next_first_volume;
	  $volume{$_} += $max_volume if(!$input{nosum});
	}
	else {
	  $volume{$_} = undef;
	}
      }
      else {
	if (!$input{nosum}) {
	  $volume{$_} += $max_volume;
	  $max_volume = $volume{$_};
	}
      }
      $highest = $volume{$x_max_value+1};
      push @links, $volume{data}->{$_};
    }

    if($input{debug}) {
      print "\%volume:\n";
      print Dumper(\%volume);
      print "\@links:\n";
      print Dumper(\@links);
    }
    $tp_count = ($max_volume - $y_min_value) / $height;
    $tp_count *= 90;
    $tp_count *= (length($max_volume) / 10);;
    $y_max_value = $max_volume + $tp_count;
    $y_tick_number = ($input{ticks})?$input{ticks}:10;
    $new1 = ($y_max_value - $y_min_value) / $y_tick_number;
    $new2 = int($new1);
    $y_max_value = int($y_min_value) + $new2 * $y_tick_number;
    $y_max_value = &roundup($y_max_value);
    if ($highest >= $y_max_value) {
    	$tp_count = 0;
    	$tp_count = ($highest - $y_min_value) / $height;
    	$tp_count *= (length($highest) / 10);
    	$tp_count *= 90;
    	$y_max_value = $highest + $tp_count;
    	$y_tick_number = ($input{ticks})?$input{ticks}:10;
	$new1 = ($y_max_value - $y_min_value) / $y_tick_number;
	$new2 = int($new1);
	$y_max_value = int($y_min_value) + $new2 * $y_tick_number;
	$y_max_value = &roundup($y_max_value);
    }
    if ($y_max_value == $y_min_value) {
    	$y_min_value = 0;
    	$y_max_value *= 1.2;
    }
    
    delete $volume{data};
    @values = map { $volume{$_} } (sort { $a <=> $b } keys %volume);

    @y_data        = (\@days, \@values);
    print "days: $#days, values: $#values\n" if($input{debug});
    print Dumper(\@y_data) . "\ny_max/y_min: $y_max_value, $y_min_value\n, $max_volume\n, $highest\n, $y_max_value2\n, new1 : $new1\n new2 : $new2\n" if($input{debug});
    $img_name = "month_" . $month . "_" . $year;
  }
  elsif ($type eq "quarter") {
    ################################# QUARTER #############################
    my($max_year, $max_month, $first_month, $last_month, $monthly_sum, @months, $currval, $i, $q_highest, $new1, $new2,$y_tick_number);
    $first_month   = ($quarter * 3) - 2;
    $last_month    = $quarter * 3;

    ($width, $height) = @size_quarter;
    if($input{width} && $input{height}) {
	$width = $input{width};
	$height = $input{height};
    }
    $graph = new GD::Graph::linespoints($width, $height);
    #$graph = new GD::Graph::lines();
    $x_tick_offset = 1;
    $x_tick_number = 0;
    $x_label_skip  = 31;

    $x_tick_length = 4;
    $x_long_ticks  = 0;
    $x_labels_vertical = 0;

    $x_label       = "$quarter. Quarter $year (" .$MonthName{$first_month} 
                   . " to " . $MonthName{$last_month} .")";
    $title         = "$maintitle";
    $show_values   = 1;


    $min_month     = $first_month - 1;
    if ($min_month == -1) {
      $min_year = $year - 1;
      $min_month = 12;
    }
    else {
      $min_year = $year;
    }

    if ($input{nosum}) {
      #$max_volume = &get_last_volume($first_month, $year);
      $y_min_value = 0;
    }
    else {
      $max_volume = &get_last_max_volume($min_month, $min_year);
      $y_min_value = $max_volume;
    }


    $max_month = $last_month + 1;
    if ($max_month == 13) {
      $max_month = 1;
      $max_year = $year + 1;
    }
    else {
      $max_year = $year;
    }
    $next_first_volume = &get_next_first_volume($max_month, $max_year);

    foreach (sort { $data{$a}->{day} <=> $data{$b}->{day} } keys %data) {
      # next if $month not between 1st and last month of current quarter!
      next if($data{$_}->{year} != $year || $data{$_}->{month} < $first_month || $data{$_}->{month} > $last_month);
      $volume{$data{$_}->{month}}->{$data{$_}->{day}} = $data{$_}->{volume};
      $volume{$data{$_}->{month}}->{data}->{$data{$_}->{day}} = $data{$_}; #store data for hotspots
      print "month: $data{$_}->{month}, day: $data{$_}->{day}, volume: $data{$_}->{volume}\n" if($input{debug});
      # i.e. %volume = ( "06" => ( 12 => 1000, 22 => 20000 ) ); !!!!!!!!!!
    }

    for my $Month ($first_month .. $last_month) {
      @days = &num_days($year, $Month);
      foreach my $Day (@days) {
	if ($Day == 1 && $Month == $first_month && $volume{$Month}->{$Day} == 0) {
	  if ($input{nosum}) {
	    push @values, 0;
	  }
	  else {
	    push @values, $max_volume;
	  }
	}
	elsif ($volume{$Month}->{$Day} == 0) {
	  push @values, undef;
	}
	else {
	  if ($input{nosum}) {
	    push @values, $volume{$Month}->{$Day};
	  }
	  else {
	    push @values, $max_volume;
	  }
	  $max_volume += $volume{$Month}->{$Day};
	}
	#$highest = $volume{$};
	push @months, $MonthName{$Month};
	push @links, $volume{$Month}->{data}->{$Day};
      }
    }

	for ($i=1;$i<$#values;$i++) {
		$currval = $values[$i];
		if ($currval > $q_highest) {
			$q_highest = $currval;
		}
	}

    $tp_count = ($q_highest - $y_min_value) / $height;
    $tp_count *= 100;
    $tp_count *= (length($q_highest) / 10);;
    $y_tick_number = ($input{ticks})?$input{ticks}:10;
    if ($tp_count != 0) {
    	$y_max_value = $q_highest + $tp_count;
	$new1 = ($y_max_value - $y_min_value) / $y_tick_number;
	$new2 = int($new1);
    	$y_max_value = int($y_min_value) + $new2 * $y_tick_number;
    	$y_max_value = &roundup($y_max_value);
    }
	if ($y_max_value == 0) {
		$y_min_value = 1;
		$y_max_value = 1;
	} elsif ($y_max_value <= $y_min_value) {
		$y_min_value = 0;
		$y_max_value = $max_volume;
		$y_max_value *= 1.2;
		$new1 = ($y_max_value - $y_min_value) / $y_tick_number;
		$new2 = int($new1);
		$y_max_value = int($y_min_value) + $new2 * $y_tick_number;
		$y_max_value = &roundup($y_max_value);
	}
	
    pop @months; #remove the last one!

    push @months, $MonthName{ $max_month };

    @y_data = (\@months, \@values);

    if($input{debug}) {
      print "\%volume:\n";
      print Dumper(\%volume);
      print "\@links:\n";
      print Dumper(\@links);
      print Dumper(\@y_data);
      print "\ny_max/y_min: $y_max_value, $y_min_value\n, $max_volume\n, $q_highest\n, $tp_count\n";
    }

    $img_name = "quarter_" . $quarter . "_" . $year;
  }
  elsif ($type eq "year") {
    ################################ YEAR #################################
    my($max_year, $max_month, $first_month, $last_month, $monthly_sum, @months, $currval, $i, $y_highest, $new1, $new2,$y_tick_number);
    ($width, $height) = @size_year;
    if($input{width} && $input{height}) {
        $width = $input{width};
        $height = $input{height};
    }

    $graph = new GD::Graph::linespoints($width, $height);
    $x_tick_number = 0;
    $x_tick_length = 4;
    $x_long_ticks  = 0;
    $x_tick_offset = 1;
    $x_label_skip  = 1; # alle 2 Monate
    $x_label       = "$year " . $MonthName{1} . " to " . $MonthName{12};
    $title         = "$maintitle";
    $show_values   = 1;
    $x_labels_vertical = 1; # long labels!


    $max_volume = &get_last_max_volume(12, $year - 1); # last years Dec max value

    if ($input{nosum}) {
      #$max_volume = &get_last_volume(12, $year - 1, 1);
      $y_min_value = 0;
    }
    else {
      $max_volume = &get_last_max_volume(12, $year - 1);
      $y_min_value = $max_volume;
    }


    $next_first_volume = &get_next_first_volume(1, $year + 1);

    foreach (sort { $data{$a}->{day} <=> $data{$b}->{day} } keys %data) {
      next if($data{$_}->{year} != $year);
      $volume{$data{$_}->{month}}->{$data{$_}->{day}} = $data{$_}->{volume};
      #$volume{$data{$_}->{month}}->{data} = $data{$_}; #store data for hotspots
      print "month: $data{$_}->{month}, day: $data{$_}->{day}, volume: $data{$_}->{volume}\n" if($input{debug});
      # i.e. %volume = ( "06" => ( 12 => 1000, 22 => 20000 ) ); !!!!!!!!!!
    }

    $first_month = 1;
    $last_month  = 12;

    for my $Month ($first_month .. $last_month) {
      @days = &num_days($year, $Month);
      foreach my $Day (@days) {
	# summarize
	$max_month += $volume{$Month}->{$Day};
      }
      if ($max_month == 0 && $Month == 1) {
	push @values, $max_volume;
      }
      elsif ($max_month == 0) {
	push @values, undef;
      }
      else {
	if ($input{nosum}) {
	  push @values, $max_month;
	}
	else {
	  push @values, $max_volume;
	}
	$max_volume += $max_month;
      }
      push @links, { month => $Month, year => $year };
      $max_month = 0;

      push @months, $MonthName{$Month};
    }


	for ($i=1;$i<12;$i++) {
		$currval = $values[$i];
		if ($currval > $y_highest) {
			$y_highest = $currval;
		}
	}


    $tp_count = ($y_highest - $y_min_value) / $height;
    $tp_count *= length($y_highest) / 10;
    $tp_count *= 130;
    $y_max_value = $y_highest + $tp_count;
    $y_tick_number = ($input{ticks})?$input{ticks}:10;
    $new1 = ($y_max_value - $y_min_value) / $y_tick_number;
    $new2 = int($new1);
    $y_max_value = int($y_min_value) + $new2 * $y_tick_number;
    $y_max_value = &roundup($y_max_value);
	if ($y_max_value == 0) {
		$y_min_value = 1;
		$y_max_value = $max_volume;
		$y_max_value *= 1.2;
		$new1 = ($y_max_value - $y_min_value) / $y_tick_number;
		$new2 = int($new1);
		$y_max_value = int($y_min_value) + $new2 * $y_tick_number;
		$y_max_value = &roundup($y_max_value);
	} elsif ($y_max_value <= $y_min_value) {
		$y_min_value = 0;
		$y_max_value = $max_volume;
		$y_max_value *= 1.2;
		$new1 = ($y_max_value - $y_min_value) / $y_tick_number;
		$new2 = int($new1);
		$y_max_value = int($y_min_value) + $new2 * $y_tick_number;
		$y_max_value = &roundup($y_max_value);
	}

    @y_data = (\@months, \@values);
    print Dumper(\@y_data) . "\ny_max, y_min: $y_max_value, $y_min_value\n $y_highest\n" if($input{debug});
    $img_name = "year_" . $year;
    #########################################################################################
  }

  if($input{debug}) {
    print "</pre>";
    &end_html;
    exit;
  }

  if ($input{nosum}) {
    $img_name .= "_nosum";
  }

  if (!exists $input{version} || $input{version} != 0) {
    $title .= "  (soldstat $Version)";
  }

  # set available colors, see rgb.txt for the corresponding names!
  GD::Graph::colour::read_rgb($RGB) or
        die "Cannot read colours from $RGB";

  #
  # generic options for GD::Graph object
  # some of them are generic for all, others are already set in the
  # if/else above
  #
  $y_tick_number = ($input{ticks})?$input{ticks}:10;
  my $ShowValues = (exists $input{values})?$input{values}:$show_values;

  $graph->set(
	      x_label_skip    => $x_label_skip,   # linie pro wieviel punkte (z.b. alle 5 Tage eine Linie!)
	      y_label_skip    => $y_label_skip,   # wie oben nur fr y-Achse (DM)
	      y_tick_number   => $y_tick_number,  # DM ticks
	      x_tick_offset   => $x_tick_offset,  # wo beim tick beginnen
	      x_label         => $x_label,
	      y_label         => "DM $input{factor}",
	      title           => $title,
	      y_max_value     => $y_max_value,
	      y_min_value     => &rounddown($y_min_value),
	      zero_axis_only  => 0,
	      long_ticks      => 1,               # ticks as long as the axes itself!
	      dclrs           => [ qw( darkorchid2 ) ], # color of the line
	      line_types      => 1,               # (1: solid, 2: dashed, 3:dotted, 4: dot-dashed)
	      line_width      => 2,
	      show_values     => $ShowValues,    # show values of points
	      bgclr           => "grey100",       #"ghost white",   # background color
	      valuesclr       => "black",         # values color
	      markers         => [ 7 ],           # filled circle
	      marker_size     => 4,               # size of the above!
	      fgclr           => "gray63",        # grid and axes color
	      values_vertical => 1,             # vertical printing
	      x_labels_vertical => $x_labels_vertical, # vertical x labels
	      values_space    => 6,
	      axislabelclr    => "mediumvioletred",
	      box_axis        => 1,
	      x_ticks         => 1,
	      x_min_value     => $x_min_value,
	      x_max_value     => $x_max_value,
	      x_tick_length   => $x_tick_length,  # len of the tick
	      x_long_ticks    => $x_long_ticks,   # 0 = short, 1 = long!
	      r_margin        => 20,
	      logo            => $LOGO,
	      logo_position   => $POSITION,
	      #types           => [ qw( points lines ) ],

	      );

  $graph->set ( x_tick_number => 'auto' )if($x_tick_number);
  $graph->set_title_font(gdSmallFont);

  $img_name .= "_" . $width . "x" . $height . "_" . $y_tick_number . "_" . $input{factor};

  # png or gif?
  my $format = $graph->export_format;

  #
  # I can put the image directly or save it
  # and then put a referencing html output
  # with image map
  #
  if ($input{return} eq "html") {
    $graph->plot(\@y_data)->$format() or die $graph->error;
    my $ext = $graph->export_format;
    
    if ($input{nosum2} eq "0") { 
    	&add_text($graph, @links); 
    }
    
    my $image = $imgpath . "/$img_name.$ext";
    open IMAGE, ">$image" || die "Cannot open $image for image writing!\n";
    binmode IMAGE;
    print IMAGE $graph->gd->$ext();
    close IMAGE;
    $image = $imghttp . "/$img_name.$ext";
    &add_text($graph, @links); 
    &respond_html($graph, $image, @links);


   }
  else {
    # hau raus, dat ding
    print CGI::header("image/$format");
    binmode STDOUT;
    print $graph->plot(\@y_data)->$format() or die $graph->error;;
    
    if ($input{nosum2} eq "0") { 
    	&add_text($graph, @links); 
    }
    
  }
}

sub add_text {
if ($input{type} ne "month") { return 0 }
my ($graph, @links) = @_;
my ($black, @pixel, $wert, $month, $year);
my $i = 0;

$year = $input{year};
$month = $input{month};

my $factor = $input{factor}; if ($factor ne "") {$factor = int($input{factor}) }
if ($factor eq undef) { $factor = 1 }

$black = $graph->gd->colorAllocate(0,0,0);
@pixel = $graph->get_hotspot;


foreach my $pos (@{$pixel[1]}) {
	$i++;
	if (!$links[$i-1]->{realvol} && $pos && $i < 3) { 
		my $lastmax = &get_real_last_volume($month - 1, $year);
		$graph->gd->stringUp (gdTinyFont, $pos->[3] + 3, $pos->[4] - 7, $lastmax / $factor . " $input{curr}", $black);
	}
	if (!$links[$i-1]->{realvol} && $pos && $i > 2) {
		my $nextmax = &get_next_first_volume($month+1, $year);
		$graph->gd->stringUp (gdTinyFont, $pos->[3] + 3, $pos->[4] - 7, $nextmax / $factor . " $input{curr}", $black);
	}
	next if(!$links[$i-1]->{realvol});
		$graph->gd->stringUp (gdTinyFont, $pos->[3] - 3, $pos->[4] - 7, $links[$i-1]->{realvol}  / $factor . " $input{curr}", $black);

}



}

sub respond_html {
  #
  # save image to file and respond with html page (includes imagemap!!)
  #
  my($graph, $image, @links) = @_;

  # get area spots from lib
  my @spots = $graph->get_hotspot(1);
  my $link;
  my $wert;

  if ($input{dumpspot}) {
    &start_html;
    print "<pre>";
    print Dumper(\@spots);
    print "</pre>";
    &end_html;
    exit;
  }

  my $i=0; # counter for @links

  # respond!
  &start_html($maintitle);
  print "<map name=\"sold\">\n";
  foreach (@spots) {
    $i++;
    next if($_ eq "");
    my($cx, $cy, $x1, $x2, $y1, $y2);
    # calculate points
    # $$_[1],$$_[4] x $$_[3],$$_[2]
    # X1     Y1       X2     Y2
    $cx = $$_[3];
    $cy = $$_[4];
    $x1 = $cx - $Xwidth;
    $x2 = $cx + $Xwidth;
    $y1 = $cy - $Ywidth;
    $y2 = $cy + $Ywidth;
    # create target link:
    if ($input{type} eq "year") {
      $link = "$baseurl/$me?state=render&factor=$input{factor}&return=html&type=month&"
	     ."month=$links[$i-1]->{month}&year=$links[$i-1]->{year}&ticks=$input{ticks}&"
	     ."volspace=$input{volspace}&nosum=$input{nosum}";
    }
    else {
      $link = "$baseurl/$me?state=showeditsingle&idx=$links[$i-1]->{idx}";
    }
    if ($1 == 1 && $links[$i-1]->{realvol} == 0) {
      print "<area shape=rect coords=\"$x1,$y1,$x2,$y2\" nohref>\n";
    }
    else {
    
      print "<area shape=rect coords=\"$x1,$y1,$x2,$y2\""
        . " title=\"$links[$i-1]->{customer}\nVolume: $links[$i-1]->{realvol} DM\" href=\"$link\">\n";
    }
  }
  print "</map>\n";
  print "<img src=\"$image\" usemap=\"#sold\" border=0>\n";
  if ($input{crlink} == 1) { &create_link }
  &end_html;
}


sub show_error {
  #
  # display nice error message
  #
  my($msg) = @_;
  &start_html;

  print "<center><table bgcolor=\"#ccccff\"border=0 cellspan=0 colspan=0>"
      . "<tr><td align=center>\n";
  print &menu . "</td><tr><td align=center>\n";
  print "<br><b><font color=red>! ERROR !</font></b><p>$msg</p><p>&nbsp;&nbsp;&nbsp;</p>";
  print "</td></tr>\n";
  print "<tr><td>" . &menu . "</td></tr></table\n";

  &end_html;
}


sub create_link {
	my $url;
	my $nosum = $input{nosum}; if ($nosum eq "1") { $nosum = "&nosum=1" }
	my $rethtml = $input{return}; if ($rethtml ne "") { $rethtml = "&return=html" }
	my $version = $input{version}; if ($version eq "0") {$version = "&version=0" }
	my $values = $input{values}; if ($values eq "0") {$values = "&values=0" }
	my $nosum2 = $input{nosum2}; if ($nosum2 eq "0") {$nosum2 = "&nosum2=0" }
	my $factor = $input{factor}; if ($factor ne "") {$factor = "&factor=$input{factor}" }
	my $month = $input{month}; if ($month ne "") {$month = "&month=$input{month}" }
	my $year = $input{year}; if ($year ne "") {$year = "&year=$input{year}" }
	my $quarter = $input{quarter}; if ($quarter ne "") {$quarter = "&quarter=$input{quarter}" }
	my $ticks = $input{ticks}; if ($ticks ne "") {$ticks = "&ticks=$input{ticks}" }
	my $width = $input{width}; if ($width ne "") {$width = "&width=$input{width}" }
	my $height = $input{height}; if ($height ne "") {$height = "&height=$input{height}" }
	my $type = $input{type}; if ($type ne "") {$type = "&type=$input{type}" }
	my $currency2 = $input{curr}; if ($currency2 ne "") {$currency2 = "&curr=$input{curr}" }
	$url = "<br>$me?state=render$values$version$nosum$rethtml$nosum2$factor$month$year$quarter$ticks$width$height$type$currency2";
	print	"\n $url \n";
}

sub num_days {
  #
  # return an array with all days
  # for a given month. the size of the array
  # tells the number of days in that month
  #
  my ($year,$month) = (@_);
  my @DAYS_IN_MONTH = qw(0 31 28 31 30 31 30 31 31 30 31 30 31);
  my ($day,@days);

  # Schaltjahr calculations
  if ((($year % 4) == 0) && ((($year % 100) != 0) || (($year % 400) == 0))) {
    $DAYS_IN_MONTH[2]++;
  }
  for($day=1; $day<=$DAYS_IN_MONTH[$month]; $day++) {
    push @days, $day;
  }
  return @days;
}






sub get_last_max_volume {
  #
  # return max volume from beginning 
  # of the database until (now -1 month)
  # for completing the curve
  #
  my($min_month, $min_year) = @_;
  my($volume, $max_volume);

  my $sql = "SELECT volume FROM $table where month <= $min_month AND year = $min_year OR year < $min_year";
  my $statement = $DB->prepare($sql) || die $DB->errstr();
  $statement->execute() || die $DB->errstr();
  $statement->bind_columns(undef, \ ($volume)) || die $DB->errstr();
  while($statement->fetch) {
    $max_volume += $volume;
  }
  if (exists $input{factor}) {
    # resize volume!
    $max_volume = int($max_volume / $input{factor}) if($input{factor} != 0);
  }
  return $max_volume;
}


sub get_last_volume {
  #
  # return the last month's volume.
  #
  my($min_month, $min_year, $day) = @_;
  my($volume, $max_volume);
  my $sql = "SELECT volume FROM $table where month = $min_month AND year = $min_year ORDER BY day";
  my $statement = $DB->prepare($sql) || die $DB->errstr();
  $statement->execute() || die $DB->errstr();
  $statement->bind_columns(undef, \ ($volume)) || die $DB->errstr();
  while($statement->fetch) {
    if (!$day) {
      $max_volume += $volume;
    }
    else {
      $max_volume = $volume;
    }
  }
  if (exists $input{factor}) {
    # resize volume!
    $max_volume = int($max_volume / $input{factor}) if($input{factor} != 0);
  }
  return $max_volume;
}

sub get_real_last_volume {

my($month, $year) = @_;
my $highest_val = 0;
my $volume;
my $sql = "SELECT volume FROM $table where month = $month AND year = $year ORDER BY day DESC";
my $statement = $DB->prepare($sql) ||die $DB->errstr();
$statement->execute() || die $DB->errstr();
$statement->bind_columns(undef, \ ($volume)) || die $DB->errstr();

while($statement->fetch) {

	if ($volume > $highest_val) { $highest_val = $volume }
}

return $highest_val;

}


sub get_next_first_volume {
  #
  # return the next months first volume
  # for completing the curve
  #
  my($min_month, $min_year) = @_;
  my($volume);
  if ($min_month == 13) {
    $min_month = 1;
    $min_year++;
  }
  my $sql = "SELECT volume FROM $table where month = $min_month AND year = $min_year ORDER by day";
  my $statement = $DB->prepare($sql) || die $DB->errstr();
  $statement->execute() || die $DB->errstr();
  $statement->bind_columns(undef, \ ($volume)) || die $DB->errstr();
  while($statement->fetch) {
    if ($volume > 0) {
      if (exists $input{factor}) {
	# resize volume!
	$volume = int($volume / $input{factor}) if($input{factor} != 0);
      }
      return $volume;
    }
  }
}





sub roundup {
  #
  # return next highest number with all zeros
  # i.e.: &roundup(192340 - 193000) will return 200000
  #
  my $zahl=shift;
  my ($var1, $var2, $firsthalf, $secondhalf);
  my $tst = length($zahl);
  if ($tst < 3) { return 0; }
  $var1 = length($zahl);
  $var1 /= 2;
  $var1 = int($var1);
  $firsthalf=substr($zahl,0,$var1);
  $firsthalf++;
  $secondhalf=substr($zahl,length($firsthalf), length($zahl) - length($firsthalf));
  my $var3 = "0"x length($secondhalf);
  $var2 = $firsthalf . $var3;
  #$zahl =~ s/^(\d)(\d+)$/($2==0?$1:($1+1)).("0"x length($2))/e;
  return $var2;
}

sub rounddown {
  #
  # return next highest number with all zeros
  # i.e.: &roundup(192340 - 193000) will return 200000
  #
  my $zahl=shift;
  my ($var1, $var2, $firsthalf, $secondhalf);
  my $tst = length($zahl);
  if ($tst < 3) { return 0; }
  $var1 = length($zahl);
  $var1 /= 2;
  $var1 = int($var1);
  $firsthalf=substr($zahl,0,$var1);
  $firsthalf--;
  $secondhalf=substr($zahl,length($firsthalf), length($zahl) - length($firsthalf));
  my $var3 = "0"x length($secondhalf);
  $var2 = $firsthalf . $var3;
  #$zahl =~ s/^(\d)(\d+)$/($2==0?$1:($1+1)).("0"x length($2))/e;
  return $var2;
}

sub sql_all {
  #
  # return all values from database
  # and add the volumes to $y_max_value
  #
  my(%data,$idx, $month, $day, $year, $time, $salesman, $customer, $product,$volume, $note, $realvol);
  my $sql = "SELECT idx, month, day, year, time, salesman, customer, product, volume, note "
    ."FROM $table";

  my $statement = $DB->prepare($sql) || die $DB->errstr();

  $statement->execute() || die $DB->errstr();

  $statement->bind_columns(undef, \ ($idx, $month, $day,
			   $year, $time, $salesman,
			   $customer, $product, $volume, $note)) || die $DB->errstr();
  while($statement->fetch) {
    $realvol = $volume;
    if (exists $input{factor}) {
      # resize volume!
      $volume = int($volume / $input{factor}) if($input{factor} != 0);
    }
    $data{$idx} = {
		 "idx"      => $idx,
		 "month"    => $month,
		 "day"      => $day,
		 "year"     => $year,
		 "time"     => $time,
		 "salesman" => $salesman,
		 "customer" => $customer,
		 "product"  => $product,
		 "realvol"  => $realvol,
		 "volume"   => $volume,
		 "note"     => $note
		};
    $y_max_value = ($volume > $y_max_value) ? $volume : $y_max_value; #increment max value on Y-axis!
  }
  return %data;
}


sub sql_all_raw {
  #
  # return all data from database as is
  # either a single item or all items depending on @_
  #
  my($single) = @_;
  my(%data,$idx, $month, $day, $year, $time, $salesman, $customer, $product,$volume, $note,);

  my $sql = "SELECT idx, month, day, year, time, salesman, customer, product, volume, note "
    ."FROM $table";

  if ($single) {
    $sql .= " WHERE idx = $single"; # get only a single one!
  }

  my $statement = $DB->prepare($sql) || die $DB->errstr();

  $statement->execute() || die $DB->errstr();

  $statement->bind_columns(undef, \ ($idx, $month, $day,
			   $year, $time, $salesman,
			   $customer, $product, $volume, $note)) || die $DB->errstr();

  while($statement->fetch) {
    if ($single) {
      %data = (
	       "idx"      => $idx,
	       "month"    => $month,
	       "day"      => $day,
	       "year"     => $year,
	       "time"     => $time,
	       "salesman" => $salesman,
	       "customer" => $customer,
	       "product"  => $product,
	       "volume"   => $volume,
	       "note"     => $note
	      );
    }
    else {
      $data{$idx} = {
		     "idx"      => $idx,
		     "month"    => $month,
		     "day"      => $day,
		     "year"     => $year,
		     "time"     => $time,
		     "salesman" => $salesman,
		     "customer" => $customer,
		     "product"  => $product,
		     "volume"   => $volume,
		     "note"     => $note
		    };
    }
  }
  return %data;
}


sub sql_connect {
  #
  # connect to the database
  #
  $database = "DBI:$dbdriver:$dbname;host=$dbhost";
  $DB = DBI->connect($database, $dbuser, $dbpasswd) || die DBI->errstr();

  $lock = $DB->prepare("LOCK TABLES $table WRITE") || die $DB->errstr();
  $lock->execute() || die $DB->errstr();
}